#include <stdlib.h>
#include <stdio.h>
#include <disp.h>
#include <dos.h>
#include <signal.h>
#include <string.h>
#include <x32.h>

#include "kernel.h"

#define NULL_STACK		512 			// Null task stack size
#define INIFL			0x200			// Initial task flags
#define DOSTICK			55				// DOS timer tick in msecs
#define DOSCOUNT		0x10000			// DOS timer divisor value

static Time_t msecs_to_run;				// Remaining time slice in msecs
static Time_t time_slice = 2 * DOSTICK;	// Full time slice in msecs
static Time_t msecs_per_tick = DOSTICK;	// Timer tick in msecs
static Time_t elapsed;					// Elapsed time in msecs
static Callback_t timer_callback;		// Timer callback

unsigned 	mt_cs,						// Application segments
			mt_ds,
			mt_es,
			mt_fs,
			mt_gs;

Task_t *mt_curr_task;					// Current task
static Task_t *last_task;				// Previous task

static Task_t main_task;				// Main task
static TaskQueue_t ready_q;				// Ready tasks queue
static TaskQueue_t terminated_q;		// Terminated tasks queue
static Switcher_t save_restore;			// Custom context switch handler

static void get_segments(void);			// Read program's segments
static void set_timer(Time_t msecs);	// Program timer frequency
static void end_mtask(void);			// Mtask clanup

// Functions for handling task states
static void block(Task_t *task, TaskState_t state);
static void ready(Task_t *task, bool success);

static void free_task(Task_t *task);	// Free task's resources

static bool select_task(void);			// Scheduling algorithm

static void free_terminated(void);		// Free terminated tasks
static void do_nothing(void *arg);		// Null task function
static bool clockint(unsigned irq);		// Timer interrupt handler

static unsigned ints_disabled;			// Interrupt disable level

typedef struct
{
	unsigned	ints_dis,
				gs,
				fs,
				es,
				ds,
				edi,
				esi,
				ebp,
				esp,
				ebx,
				edx,
				ecx,
				eax,
				eflags;	
}
Regs_t;									// Stack frame for context switch

typedef struct
{
	Regs_t		regs;
	void		(*func)(void *arg);
	void		(*killer)(void);
	void *		arg;
}
InitialStack_t;							// Initial task stack frame

/*
--------------------------------------------------------------------------------
GetTime - get the current elapsed time in milliseconds
--------------------------------------------------------------------------------
*/

Time_t
GetTime(void)
{
	return elapsed;
}

/*
--------------------------------------------------------------------------------
Malloc, Realloc, StrDup, Free - heap handling functions
--------------------------------------------------------------------------------
*/

void *
Malloc(unsigned size)
{
	void *p;

	Atomic();
	free_terminated();
	if ( !(p = malloc(size)) )
		Panic("Malloc error");
	memset(p, 0, size);
	Unatomic();
	return p;
}

void *
Realloc(void *mem, unsigned size)
{
	void *p;

	if ( !mem )
		return Malloc(size);
	Atomic();
	free_terminated();
	if ( !(p = realloc(mem, size)) )
		Panic("Realloc error");
	Unatomic();
	return p;
}

char *
StrDup(char *str)
{
	char *p;

	if ( !str )
		return NULL;
	Atomic();
	free_terminated();
	if ( !(p = strdup(str)) )
		Panic("Strdup error");
	Unatomic();
	return p;
}

void
Free(void *mem)
{
	if ( !mem )
		return;
	Atomic();
	free(mem);
	Unatomic();
}

/*
--------------------------------------------------------------------------------
block - blocks a task

Tha task is removed from any task queue and its state is set.
--------------------------------------------------------------------------------
*/

static void
block(Task_t *task, TaskState_t state)
{
	mt_dequeue(task);
	mt_dequeue_time(task);
	task->state = state;
}

/*
--------------------------------------------------------------------------------
ready - unblocks a task

Sets the task's state to ready and puts the task in the ready queue. If the
task was blocked in WaitQueue, Send or Receive, the success argument becomes
the return status of the blocking function.
--------------------------------------------------------------------------------
*/

static void
ready(Task_t *task, bool success)
{
	if ( task->state == TaskReady )
		return;

	mt_dequeue(task);
	mt_dequeue_time(task);
	mt_enqueue(task, &ready_q);
	task->success = success;
	task->state = TaskReady;
}

/*
--------------------------------------------------------------------------------
CreateTask - make a new task

This function takes a pointer to the task's body, of type void f(void*), stack
size, a pointer to the task's arguments and the initial taks's priority.
The taks's stack is allocated and an initial stack frame is built. The task's 
return address is set to Exit(), so that it is automatically killed if it 
returns from its function. The task is initially suspended; it begins running
when the application calls Ready() on it.
--------------------------------------------------------------------------------
*/

Task_t *
CreateTask(TaskFunc_t func, unsigned stacksize, void *arg,
						char *name, Prio_t priority)
{
	Task_t *task;
	InitialStack_t *s;

	// adjust stack size to be multiple of 4
	stacksize = quad(stacksize);

	// allocate task control block and stack
	task = Malloc(sizeof(Task_t));
	task->name = task->send_queue.name = StrDup(name);
	task->priority = priority;
	task->stack = Malloc(stacksize);

	// setup initial stack
	s = (InitialStack_t *)(task->stack + stacksize) - 1;
	task->ss = mt_ds;
	task->esp = (unsigned) s;
	s->func = func;
	s->arg = arg;
	s->killer = Exit;				// func()'s return address

	// setup initial context
	s->regs.eflags = INIFL;
	s->regs.ints_dis = 0;
	s->regs.ds = mt_ds;
	s->regs.es = mt_es;
	s->regs.fs = mt_fs;
	s->regs.gs = mt_gs;

	return task;
}

/*
--------------------------------------------------------------------------------
ProtectMath - initialize math context for a task

To be used on all tasks using the 80387 (if more that one).
--------------------------------------------------------------------------------
*/

void
ProtectMath(Task_t *task)
{
	if ( !task->math_data )
		mt_init_math(task);
}

/*
--------------------------------------------------------------------------------
DeleteTask - delete a task created by CreateTask

If called on the current task resources cannot be freed immediately, so the
task is put in the terminated tasks queue for later cleanup.
--------------------------------------------------------------------------------
*/

static void
free_task(Task_t *task)
{
	free(task->name);
	free(task->stack);
	if ( task->math_data )
		free(task->math_data);
	free(task);
}

void
DeleteTask(Task_t *task)
{
	if ( task == &main_task )
		Panic("Cannot delete main task");

	FlushQueue(&task->send_queue, false);
	DisableInts();
	if ( task == mt_curr_task )
	{
		mt_curr_task->state = TaskTerminated;
		mt_enqueue(mt_curr_task, &terminated_q);
		Scheduler();
	}
	else
	{
		block(task, TaskTerminated);
		free_task(task);
	}
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
free_terminated - free terminated tasks' resources
--------------------------------------------------------------------------------
*/

void
free_terminated(void)
{
	Task_t *task;

	while ( (task = mt_getlast(&terminated_q)) )
		free_task(task);
}

/*
--------------------------------------------------------------------------------
CurrentTask - return a pointer to the current task control block
--------------------------------------------------------------------------------
*/

Task_t *
CurrentTask(void)
{
	return mt_curr_task;
}


/*
--------------------------------------------------------------------------------
Panic - unrecoverable system error
--------------------------------------------------------------------------------
*/

void
Panic(char *msg)
{
	Atomic();
	disp_printf("\r\n[%s]: %s\r\n", mt_curr_task->name, msg);
	exit(1);
}

/*
--------------------------------------------------------------------------------
Pause - suspend current task
--------------------------------------------------------------------------------
*/

void
Pause(void)
{
	Suspend(mt_curr_task);
}

/*
--------------------------------------------------------------------------------
Yield - voluntarily relinquish CPU

This will give the CPU to any task with same or higher priority than the calling
task.
--------------------------------------------------------------------------------
*/

void
Yield(void)
{
	Ready(mt_curr_task);
}

/*
--------------------------------------------------------------------------------
Delay - put the current task to sleep for a given time interval in milliseconds
--------------------------------------------------------------------------------
*/

void
Delay(Time_t msecs)
{
	DisableInts();
	if ( msecs )
	{
		block(mt_curr_task, TaskDelaying);
		if ( msecs != FOREVER )
			mt_enqueue_time(mt_curr_task, msecs);
	}
	else
		ready(mt_curr_task, false);
	Scheduler();
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
DelayUntil - delay until some time in the future

Does not block if the deadline has already expired.
--------------------------------------------------------------------------------
*/

void
DelayUntil(Time_t deadline)
{
	int delay;

	DisableInts();
	if ( (delay = (int) deadline - (int) elapsed) >= 0 )
		Delay((Time_t) delay);
	RestoreInts();
}


/*
--------------------------------------------------------------------------------
Exit - end the current task

All tasks created with CreateTask will automatically call Exit() when
returning from the task's function. Exit() never returns.
--------------------------------------------------------------------------------
*/

void
Exit(void)
{
	DeleteTask(mt_curr_task);
}

/*
--------------------------------------------------------------------------------
Suspend - suspend a task
--------------------------------------------------------------------------------
*/

void
Suspend(Task_t *task)
{
	DisableInts();
	block(task, TaskSuspended);
	if ( task == mt_curr_task )
		Scheduler();
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
Ready - put a task in the ready queue

Causes a failure return status to a task sleeping in a blocking call.
--------------------------------------------------------------------------------
*/

void
Ready(Task_t *task)
{
	DisableInts();
	ready(task, false);
	Scheduler();
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
GetPriority - return a task's priority
--------------------------------------------------------------------------------
*/

Prio_t
GetPriority(Task_t *task)
{
	return task->priority;
}

/*
--------------------------------------------------------------------------------
SetPriority - set a task's priority

If the task is currently in a task queue, it is removed and re-inserted to
reflect the new priority. When changing the priority of the current task or
of a ready task, the scheduler is called to re-evaluate CPU assignment.
--------------------------------------------------------------------------------
*/

void		
SetPriority(Task_t *task, Prio_t priority)
{
	TaskQueue_t *queue;

	DisableInts();
	task->priority = priority;
	if ( (queue = task->queue) )
	{
		mt_dequeue(task);
		mt_enqueue(task, queue);
	}
	if ( task == mt_curr_task || task->state == TaskReady )
		Scheduler();
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
SetData - set a pointer to a taks's private data
--------------------------------------------------------------------------------
*/

void		
SetData(Task_t *task, void *data)
{
	DisableInts();
	task->data = data;
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
SetSwitcher - set a global context switch handler

This is function will be called on each task switch. It is passed pointers to
the previous task, whose context must be saved, and to the current task, whose
context must be restored. Interrupts are disabled and MUST NOT be enabled.
--------------------------------------------------------------------------------
*/

void		
SetSwitcher(Switcher_t switcher)
{
	DisableInts();
	save_restore = switcher;
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
CreateQueue - create a task queue
--------------------------------------------------------------------------------
*/

TaskQueue_t *	
CreateQueue(char *name)
{
	TaskQueue_t *queue = Malloc(sizeof(TaskQueue_t));

	queue->name = StrDup(name);
	return queue;
}

/*
--------------------------------------------------------------------------------
DeleteQueue - destroy a task queue
--------------------------------------------------------------------------------
*/

void
DeleteQueue(TaskQueue_t *queue)
{
	FlushQueue(queue, false);
	Free(queue->name);
	Free(queue);
}

/*
--------------------------------------------------------------------------------
WaitQueue, WaitQueueTimed - wait in a task queue

This function returns true if the task was awakened by SignalQueue. If the
task was awakened by FlushQueue(), the return status is FlushQueue()'s second
argument.
--------------------------------------------------------------------------------
*/

bool			
WaitQueue(TaskQueue_t *queue)
{
	return WaitQueueTimed(queue, FOREVER);
}

bool			
WaitQueueTimed(TaskQueue_t *queue, Time_t msecs)
{
	bool success;

	if ( !msecs )
		return false;

	DisableInts();
	block(mt_curr_task, TaskWaiting);
	mt_enqueue(mt_curr_task, queue);
	if ( msecs != FOREVER )
		mt_enqueue_time(mt_curr_task, msecs);
	Scheduler();
	success = mt_curr_task->success;
	RestoreInts();

	return success;
}

/*
--------------------------------------------------------------------------------
SignalQueue, FlushQueue - wake up tasks waiting in a task queue

SignalQueue() wakes up the last task in the queue (the task with higher
priority) and returns true, or returns false if the queue was empty. The awakened
task completes its WaitQueue() successfully.
FlushQueue() wakes up all tasks in the queue, which complete their respective
WaitQueue()s with a success or failure status, depending on FlusQueue()'s second
argument.
--------------------------------------------------------------------------------
*/

bool		
SignalQueue(TaskQueue_t *queue)
{
	Task_t *task;

	DisableInts();
	if ( (task = mt_getlast(queue)) )
	{
		ready(task, true);
		Scheduler();
	}
	RestoreInts();

	return task != NULL;
}

void			
FlushQueue(TaskQueue_t *queue, bool success)
{
	Task_t *task;

	DisableInts();
	if ( mt_peeklast(queue) )
	{
		while ( (task = mt_getlast(queue)) )
			ready(task, success);
		Scheduler();
	}
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
Send, SendCond, SendTimed - send a message to another task

Argument usage:

	msg == NULL		-> No message passed, just wake up sender && destination
	msg != NULL		-> Pass size bytes to destination task
--------------------------------------------------------------------------------
*/

bool			
Send(Task_t *to, void *msg, unsigned size)
{
	return SendTimed(to, msg, size, FOREVER);
}

bool			
SendCond(Task_t *to, void *msg, unsigned size)
{
	return SendTimed(to, msg, size, 0);
}

bool			
SendTimed(Task_t *to, void *msg, unsigned size, Time_t msecs)
{
	bool success;

	DisableInts();

	if ( to->state == TaskReceiving && (!to->from || to->from == mt_curr_task) )
	{
		to->from = mt_curr_task;
		if ( to->msg && msg )
		{
			if ( size > to->size )
				Panic("Receiving buffer too small");
			to->size = size;
			memcpy(to->msg, msg, size);
		}
		else
			to->size = 0;
		ready(to, true);
		Scheduler();
		RestoreInts();
		return true;
	}

	if ( !msecs )
	{
		RestoreInts();
		return false;
	}

	mt_curr_task->msg = msg;
	mt_curr_task->size = size;
	mt_curr_task->state = TaskSending;
	mt_enqueue(mt_curr_task, &to->send_queue);
	if ( msecs != FOREVER )
		mt_enqueue_time(mt_curr_task, msecs);
	Scheduler();
	success = mt_curr_task->success;

	RestoreInts();
	return success;
}


/*
--------------------------------------------------------------------------------
Receive, ReceiveCond, ReceiveTimed - receive a message

Argument usage:

	 from == NULL		-> receive from any
	*from == NULL		-> receive from any, get sender address in *from
	*from == sometask	-> receive from sometask only.

	 msg == NULL	-> no message passed, just wake up sender && destination
	 msg != NULL	-> get message in buffer of size *size, get length of 
	 				   message in *size.
--------------------------------------------------------------------------------
*/

bool			
Receive(Task_t **from, void *msg, unsigned *size)
{
	return ReceiveTimed(from, msg, size, FOREVER);
}

bool			
ReceiveCond(Task_t **from, void *msg, unsigned *size)
{
	return ReceiveTimed(from, msg, size, 0);
}

bool			
ReceiveTimed(Task_t **from, void *msg, unsigned *size, Time_t msecs)
{
	bool success;
	Task_t *sender;

	DisableInts();

	if ( from && *from )
		sender = (*from)->queue == &mt_curr_task->send_queue ? *from : NULL;
	else
		sender = mt_peeklast(&mt_curr_task->send_queue);

	if ( sender )
	{
		if ( from ) 
			*from = sender;
		if ( sender->msg && msg )
		{
			if ( size )
			{
				if ( sender->size > *size )
					Panic("Receiving buffer too small");
				memcpy(msg, sender->msg, *size = sender->size);
			}
		}
		else if ( size )
			*size = 0;
		ready(sender, true);
		Scheduler();
		RestoreInts();
		return true;
	}

	if ( !msecs )
	{
		RestoreInts();
		return false;
	}

	mt_curr_task->from = from ? *from : NULL;
	mt_curr_task->msg = msg;
	mt_curr_task->size = size ? *size : 0;
	mt_curr_task->state = TaskReceiving;
	if ( msecs != FOREVER )
		mt_enqueue_time(mt_curr_task, msecs);
	Scheduler();
	if ( (success = mt_curr_task->success) )
	{
		if ( size )
			*size = mt_curr_task->size;
		if ( from )
			*from = mt_curr_task->from;
	}

	RestoreInts();
	return success;
}

/*
--------------------------------------------------------------------------------
select_task - select the next task to run

Return value: whether the current task (mt_curr_task) has been changed. If it
has, the previously running task is stored in last_task .This function is
called from Scheduler(). It takes care of saving/restoring the math context
if applicable, and calling the user installed context switch handler if
installed.
--------------------------------------------------------------------------------
*/

static bool 
select_task(void)
{
	Task_t *ready_task;

	// Check if the current task can continue running
	if ( mt_curr_task->state == TaskCurrent )
	{
		if ( mt_curr_task->atomic_level )		// don't disturb
			return false;

		// Check priorities and time slice
		ready_task = mt_peeklast(&ready_q);
		if ( !ready_task || ready_task->priority < mt_curr_task->priority ||
			(msecs_to_run && ready_task->priority == mt_curr_task->priority) )
			return false; 

		// Current task should be preempted
		ready(mt_curr_task, false);
	}

	// Get next task
	last_task = mt_curr_task;
	mt_curr_task = mt_getlast(&ready_q);
	mt_curr_task->state = TaskCurrent;

	// If it is the same, nothing else to do
	if ( mt_curr_task == last_task )
		return false;

	// Save/restore math context if applicable
	if ( last_task->math_data )
		mt_save_math(last_task);
	if ( mt_curr_task->math_data )
		mt_restore_math(mt_curr_task);

	// Save/restore user context
	if ( save_restore )
		save_restore(last_task, mt_curr_task);

	//Initialize time slice
	msecs_to_run = time_slice;
	return true;
}

/*
--------------------------------------------------------------------------------
Scheduler - select next task to run and eventually change context

This is called when the current task blocks, or any task is awakened, or the
priority of current or ready tasks are changed, or when returning from a 
first-level (non-nested) interrupt.
If called from with an interrupt handler (mt_int_level != 0) it does not do
anything, since rescheduling is delayed until returning from the first level
of interrupts. This happens, for example, if a signal handler calls
SignalQueue() or SendCond() to wake up a sleeping task.
--------------------------------------------------------------------------------
*/

__declspec(naked)
void
Scheduler(void)
{
	asm
	{
		cli
		pushfd								// push context
		pushad	
		push	ds
		push	es
		push	fs
		push	gs
		mov		ds, word ptr cs:mt_ds
		push	ints_disabled

		cmp		mt_int_level, 0				// interrupt handler, don't
		jne		restore						// reschedule

		call	select_task					// reschedule
		test	eax, eax					// check if current task changed
		jz		restore

		mov		eax, last_task				// save previous task's stack ptr
		mov		[eax]Task_t.esp, esp
		mov		word ptr [eax]Task_t.ss, ss

		mov		eax, mt_curr_task			// switch to current task's stack
		lss		esp, qword ptr [eax]Task_t.esp

	restore:
		pop		ints_disabled				// pop context
		cmp		ints_disabled, 0
		pop		gs
		pop		fs
		pop		es
		pop		ds
		popad
		jne		ints_blocked		
		popfd
		sti
		ret
	ints_blocked:
		popfd
		cli
		ret
	}
}

/*
--------------------------------------------------------------------------------
SetTimerCallback - set a callback to be called on every timer interrupt.
--------------------------------------------------------------------------------
*/

void
SetTimerCallback(Callback_t callback)
{
	timer_callback = callback;
}

/*
--------------------------------------------------------------------------------
clockint - timer interrupt

Wake up tasks in the time queue with waiting time less than one timer tick. 
Decrease current task's time slice by the amount of a timer tick. 
Call timer callback if installed.
Chain to original DOS handler every DOSTICK milliseconds (on the average).
--------------------------------------------------------------------------------
*/

static bool 
clockint(unsigned irq)
{
	static Time_t dos_count;
	Task_t *task;

	elapsed += msecs_per_tick;

	if ( msecs_to_run > msecs_per_tick )
		msecs_to_run -= msecs_per_tick;
	else
		msecs_to_run = 0;

	while ( (task = mt_peekfirst_time()) && task->msecs <= msecs_per_tick )
	{
		mt_getfirst_time();
		ready(task, false);
	}
	if ( task )
		task->msecs -= msecs_per_tick;

	if ( timer_callback )
		timer_callback();

	if ( (dos_count += msecs_per_tick) < DOSTICK )
	{
		EndInterrupt();
		return true;			// don't chain
	}
	dos_count -= DOSTICK;
	return false;				// chain to DOS handler
}

/*
--------------------------------------------------------------------------------
Atomic, Unatomic - disable/restore preemptions for the current task

These functions may be nested; preemptions are enabled when the disabling
counter is zero. When re-enabling preemptions the scheduler is called.
--------------------------------------------------------------------------------
*/

void
Atomic(void)
{
	++mt_curr_task->atomic_level;
}

void
Unatomic(void)
{
	if ( mt_curr_task->atomic_level && !--mt_curr_task->atomic_level )
		Scheduler();
}

/*
--------------------------------------------------------------------------------
DisableInts, RestoreInts - disable/restore interrupts for the current task

These functions may be nested; interrupts are enabled when the disabling
counter is zero.
--------------------------------------------------------------------------------
*/

void
DisableInts(void)
{
	disable();
	ints_disabled++;
}

void
RestoreInts(void)
{
	if ( ints_disabled && !--ints_disabled )
		enable();
}

/*
--------------------------------------------------------------------------------
do_nothing - null task

Runs with priority 0, using the CPU when no other task es ready.
--------------------------------------------------------------------------------
*/

static void
do_nothing(void *arg)
{
	while ( true )
		;
}


/*
--------------------------------------------------------------------------------
get_segments - get program's segments
--------------------------------------------------------------------------------
*/

static void
get_segments(void)
{
	asm
	{
		mov		word ptr mt_cs, cs
		mov		word ptr mt_ds, ds
		mov		word ptr mt_es, es
		mov		word ptr mt_fs, fs
		mov		word ptr mt_gs, gs
	}
}


/*
--------------------------------------------------------------------------------
set_timer - program real time clock frequency
--------------------------------------------------------------------------------
*/

static void
set_timer(Time_t msecs)
{
	if ( !msecs || msecs > DOSTICK )
		Panic("set_timer: illegal timer interval");

	msecs *= DOSCOUNT;
	msecs /= DOSTICK;
	disable();
	_outp(0x43, 0x36);
	_outp(0x40, msecs);
	_outp(0x40, msecs >> 8);
	enable();
}


/*
--------------------------------------------------------------------------------
InitMTask, end_mtask - Mtask initialization and cleanup

InitMtask() creates the main and null tasks and hooks the timer interrupt.
The main task works like any other task, but does not have a dynamically
allocated stack. Instead, it uses the program's original stack. The main task
cannot be destroyed by calling DeleteTask() or Exit(), and when main() returns
or calls exit() the whole program ends.
Null values for mspertick and timeslice default to DOSTICK milliseconds and
two ticks respectively.
--------------------------------------------------------------------------------
*/

void
InitMTask(Time_t mspertick, Time_t timeslice)
{
	// Get program segments
	get_segments();

	// Open display
	disp_open();

	// Set time slice
	if ( timeslice )
		time_slice = timeslice;
	
	// Initialize main task
	main_task.name = "Main Task";
	main_task.state = TaskCurrent;
	main_task.priority = DEFAULT_PRIO;
	main_task.send_queue.name = main_task.name;
	msecs_to_run = time_slice;
	mt_curr_task = &main_task;

	// Initialize null task
	ready(CreateTask(do_nothing, NULL_STACK, NULL, "Null Task", MIN_PRIO), 
			false);

	// Initialize math module
	mt_setup_math();

	// Schedule cleanup upon exit
	atexit(end_mtask);
	
	// Set timer frequency
	if ( mspertick )
		set_timer(msecs_per_tick = mspertick);

	// Initialize interrupt handling module and hook timer interrupt
	mt_setup_irqs();
	SetHandler(CLOCKIRQ, clockint);
}

static void
end_mtask(void)
{
	Atomic();
	// Restore DOS timer freq.
	if ( msecs_per_tick != DOSTICK )
		set_timer(DOSTICK);
	// Restore interrupt system
	mt_restore_irqs();
}
