#include <stdlib.h>
#include <stdio.h>

#include "pipe.h"

#define min(a, b) (((a) < (b)) ? (a) : (b))

/*
--------------------------------------------------------------------------------
get_data, put_data - internal functions for getting/putting data

These functions work on a circular byte buffer. Put_data() puts as many of
the required bytes as it can and returns the number of bytes put, or blocks on
a condition variable if the buffer is initially full. Conversely, get_data()
gets as many of the required bytes as it can, or blocks on a condition variable
if the buffer is initially empty.
--------------------------------------------------------------------------------
*/

static unsigned
get_data(Pipe_t *p, char *data, unsigned nreq, Time_t msecs)
{
	unsigned i, nbytes;

	EnterMonitor(p->monitor);
	if ( !p->avail && !WaitConditionTimed(p->cond_get, msecs) )
	{
		LeaveMonitor(p->monitor);
		return 0;
	}
	for ( nbytes = min(nreq, p->avail), i = 0 ; i < nbytes ; i++ )
	{
		*data++ = *p->head++;
		if ( p->head == p->end )
			p->head = p->buf;
	}
	p->avail -= nbytes;
	SignalCondition(p->cond_put);
	LeaveMonitor(p->monitor);
	return nbytes;
}

static unsigned
put_data(Pipe_t *p, char *data, unsigned nreq, Time_t msecs)
{
	unsigned i, nbytes;

	if ( !nreq )
		return 0;

	EnterMonitor(p->monitor);
	if ( p->avail == p->size && !WaitConditionTimed(p->cond_put, msecs) )
	{
		LeaveMonitor(p->monitor);
		return 0;
	}
	for ( nbytes = min(nreq, p->size - p->avail), i = 0 ; i < nbytes ; i++ )
	{
		*p->tail++ = *data++;
		if ( p->tail == p->end )
			p->tail = p->buf;
	}
	p->avail += nbytes;
	SignalCondition(p->cond_get);
	LeaveMonitor(p->monitor);
	return nbytes;
}

/*
--------------------------------------------------------------------------------
CreatePipe, DeletePipe - create/destroy a pipe

The size parameter sets the size of the internal buffer in bytes. This value
is arbitrary and may be as small as 1. Using a small buffer increases the 
coupling between producers and consumers, and forces them to run in lockstep.
Increasing the size of the buffer de-couples producers and consumers. The two 
booleans indicate if input/output should be serialized by mutexes, in case there
are several producers/consumers.
--------------------------------------------------------------------------------
*/

Pipe_t *
CreatePipe(char *name, unsigned size, bool serialized_get, bool serialized_put)
{
	char buf[200];
	Pipe_t *p = Malloc(sizeof(Pipe_t));

	p->head = p->tail = p->buf = Malloc(p->size = size);
	p->end = p->buf + size;
	p->monitor = CreateMonitor(name);
	sprintf(buf, "Get %s", name);
	if ( serialized_get )
		p->mutex_get = CreateMutex(buf);
	p->cond_get = CreateCondition(buf, p->monitor);
	sprintf(buf, "Put %s", name);
	if ( serialized_put )
		p->mutex_put = CreateMutex(buf);
	p->cond_put = CreateCondition(buf, p->monitor);

	return p;
}

void
DeletePipe(Pipe_t *p)
{
	if ( p->mutex_get )
		DeleteMutex(p->mutex_get);
	DeleteCondition(p->cond_get);
	if ( p->mutex_put )
		DeleteMutex(p->mutex_put);
	DeleteCondition(p->cond_put);
	DeleteMonitor(p->monitor);
	Free(p->buf);
	Free(p);
}

/*
--------------------------------------------------------------------------------
GetPipe, GetPipeCond, GetPipeTimed - read from a pipe

The functions try to read size bytes from the pipe, and return the quantity 
of bytes read. The blocking behaviour depends on the get_all parameter:

 	get_all == true		GetPipe() blocks until size bytes read.
 						GetPipeTimed() blocks until size bytes read or timeout.
 						GetPipeCond() returns immediately with whatever is
 							available,or zero.
 	get_all == false	GetPipe() blocks until something is available
 							(even if less that size bytes).
 						GetPipeTimed() blocks until something is available 
 							(even if less than size bytes) or timeout.
 						GetPipeCond() returns immediately with whatever is
 							available,or zero.

If serialized, reads from the pipe are atomic.
--------------------------------------------------------------------------------
*/

unsigned
GetPipe(Pipe_t *p, void *data, unsigned size, bool get_all)
{
	return GetPipeTimed(p, data, size, FOREVER, get_all);
}

unsigned
GetPipeCond(Pipe_t *p, void *data, unsigned size)
{
	return GetPipeTimed(p, data, size, 0, false);
}

unsigned
GetPipeTimed(Pipe_t *p, void *data, unsigned size, Time_t msecs, bool get_all)
{
	int n, nbytes;
	char *d;

	if ( !size || (p->mutex_get && !EnterMutexTimed(p->mutex_get, msecs)) )
		return 0;
	
	if ( get_all )
		for ( d = data, nbytes = 0 ; nbytes < size &&
				(n = get_data(p, d+nbytes, size-nbytes, msecs)) ; nbytes += n )
			;
	else
		nbytes = get_data(p, data, size, msecs);

	if ( p->mutex_get )
		LeaveMutex(p->mutex_get);
	return nbytes;
}

/*
--------------------------------------------------------------------------------
PutPipe, PutPipeCond, PutPipeTimed - write to pipe.

The functions try to write size bytes int the pipe, and return the quantity 
of bytes written.

 	PutPipe() blocks until size bytes written.
 	PutPipeTimed() blocks until size bytes written or timeout.
 	PutPipeCond() writes as much as it can and returns immediately.

If serialized, writes to the pipe are atomic.
--------------------------------------------------------------------------------
*/

unsigned
PutPipe(Pipe_t *p, void *data, unsigned size)
{
	return PutPipeTimed(p, data, size, FOREVER);
}

unsigned
PutPipeCond(Pipe_t *p, void *data, unsigned size)
{
	return PutPipeTimed(p, data, size, 0);
}

unsigned
PutPipeTimed(Pipe_t *p, void *data, unsigned size, Time_t msecs)
{
	int n, nbytes;
	char *d;

	if ( !size || (p->mutex_put && !EnterMutexTimed(p->mutex_put, msecs)) )
		return 0;
	
	for ( d = data, nbytes = 0 ; nbytes < size &&
			(n = put_data(p, d+nbytes, size-nbytes, msecs)) ; nbytes += n )
		;

	if ( p->mutex_put )
		LeaveMutex(p->mutex_put);
	return nbytes;
}

/*
--------------------------------------------------------------------------------
AvailPipe - tells how many bytes are stored in the pipe.
--------------------------------------------------------------------------------
*/

unsigned
AvailPipe(Pipe_t *p)
{
	return p->avail;
}
