#include "sem.h"

/*
--------------------------------------------------------------------------------
CreateSem - create a semaphore with an initial count
--------------------------------------------------------------------------------
*/

Semaphore_t *CreateSem(char *name, unsigned value)
{
	Semaphore_t *sem = Malloc(sizeof(Semaphore_t));

	sem->queue = CreateQueue(name);
	sem->value = value;
	return sem;
}

/*
--------------------------------------------------------------------------------
DeleteSem - destroy a semaphore
--------------------------------------------------------------------------------
*/

void
DeleteSem(Semaphore_t *sem)
{
	DeleteQueue(sem->queue);
	Free(sem);
}

/*
--------------------------------------------------------------------------------
WaitSem, WaitSemCond, WaitSemTimed - wait on a semaphore

If the semaphore's count is greater than zero, decrement it and return true.
Otherwise block on the semaphore's task queue and return true if the task was
awakened because the semaphore was signalled, false otherwise.
--------------------------------------------------------------------------------
*/

bool
WaitSem(Semaphore_t *sem)
{
	return WaitSemTimed(sem, FOREVER);
}

bool
WaitSemCond(Semaphore_t *sem)
{
	return WaitSemTimed(sem, 0);
}

bool
WaitSemTimed(Semaphore_t *sem, Time_t msecs)
{
	bool success;

	DisableInts();
	if ( (success = (sem->value > 0)) )
		sem->value--;
	else
		success = WaitQueueTimed(sem->queue, msecs);
	RestoreInts();

	return success;
}

/*
--------------------------------------------------------------------------------
SignalSem - signal a semaphore

Wakes up the first task waiting in the semaphore's task queue, which then
successfully completes its WaitSem() operation, or increments the semaphore's
count if the queue is empty.
--------------------------------------------------------------------------------
*/

void
SignalSem(Semaphore_t *sem)
{
	DisableInts();
	if ( !SignalQueue(sem->queue) )
		sem->value++;
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
ValueSem - tells the semaphore's count
--------------------------------------------------------------------------------
*/

unsigned	
ValueSem(Semaphore_t *sem)
{
	return sem->value;
}

/*
--------------------------------------------------------------------------------
FlushSem - wakes up all taks waiting on a semaphore.

The awakened tasks complete their WaitSem() operations with the wait_ok return
status.
--------------------------------------------------------------------------------
*/

void
FlushSem(Semaphore_t *sem, bool wait_ok)
{
	DisableInts();
	sem->value = 0;
	FlushQueue(sem->queue, wait_ok);
	RestoreInts();
}
